//--------------------------------------------------------
// Multiply 4x4 Matrix
//   2013/02/02 Jun Mizutani
// as -mcpu=arm1176jzf-s -mfpu=vfp mat4x4.s -o mat4x4asm.o
// gcc -Wall -o mat4 mat4.c mat4x4asm.o
// gcc -O2 -Wall -o mat4-O2 mat4.c mat4x4asm.o
//--------------------------------------------------------

#include <stdio.h>
#include <sys/time.h>

typedef struct Matrix4 {
    double m0,m1,m2,m3, m4,m5,m6,m7, m8,m9,m10,m11, m12,m13,m14,m15;
} Matrix4;

Matrix4 a = { 1, 2, 3, 4,  1, 2, 3, 4,  1, 2, 3, 4,     1, 2, 3, 4 };
Matrix4 b = { 0, 1, 2, 3,  4, 5, 6, 7,  8, 9, 10, 11,  12, 13, 14, 15 };
Matrix4 c = { 0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,     0, 0, 0, 0 };

extern void MulMatrix4x4asm(Matrix4 *result, Matrix4 *a, Matrix4 *b);

void MulMatrix4(Matrix4 *a, Matrix4 *b, Matrix4 *c)
{
    c->m0 = a->m0 * b->m0 + a->m4 * b->m1 + a->m8 * b->m2 + a->m12 * b->m3;
    c->m1 = a->m1 * b->m0 + a->m5 * b->m1 + a->m9 * b->m2 + a->m13 * b->m3;
    c->m2 = a->m2 * b->m0 + a->m6 * b->m1 + a->m10* b->m2 + a->m14 * b->m3;
    c->m3 = a->m3 * b->m0 + a->m7 * b->m1 + a->m11* b->m2 + a->m15 * b->m3;

    c->m4 = a->m0 * b->m4 + a->m4 * b->m5 + a->m8 * b->m6 + a->m12 * b->m7;
    c->m5 = a->m1 * b->m4 + a->m5 * b->m5 + a->m9 * b->m6 + a->m13 * b->m7;
    c->m6 = a->m2 * b->m4 + a->m6 * b->m5 + a->m10* b->m6 + a->m14 * b->m7;
    c->m7 = a->m3 * b->m4 + a->m7 * b->m5 + a->m11* b->m6 + a->m15 * b->m7;

    c->m8 = a->m0 * b->m8 + a->m4 * b->m9 + a->m8 * b->m10+ a->m12 * b->m11;
    c->m9 = a->m1 * b->m8 + a->m5 * b->m9 + a->m9 * b->m10+ a->m13 * b->m11;
    c->m10= a->m2 * b->m8 + a->m6 * b->m9 + a->m10* b->m10+ a->m14 * b->m11;
    c->m11= a->m3 * b->m8 + a->m7 * b->m9 + a->m11* b->m10+ a->m15 * b->m11;

    c->m12= a->m0 * b->m12+ a->m4 * b->m13+ a->m8 * b->m14+ a->m12 * b->m15;
    c->m13= a->m1 * b->m12+ a->m5 * b->m13+ a->m9 * b->m14+ a->m13 * b->m15;
    c->m14= a->m2 * b->m12+ a->m6 * b->m13+ a->m10* b->m14+ a->m14 * b->m15;
    c->m15= a->m3 * b->m12+ a->m7 * b->m13+ a->m11* b->m14+ a->m15 * b->m15;
}

void PrintMatrix4(Matrix4 *a)
{
    printf("%10.5e %10.5e %10.5e %10.5e \n", a->m0, a->m4, a->m8,  a->m12);
    printf("%10.5e %10.5e %10.5e %10.5e \n", a->m1, a->m5, a->m9,  a->m13);
    printf("%10.5e %10.5e %10.5e %10.5e \n", a->m2, a->m6, a->m10, a->m14);
    printf("%10.5e %10.5e %10.5e %10.5e \n\n", a->m3, a->m7, a->m11, a->m15);
}

int main(int argc, char* argv[])
{
  struct timeval start, now;
  struct timezone tzone;
  double elapsedTime = 0.0;
  int i;

  printf("start C version\n");
  gettimeofday(&start, &tzone);

  for(i=0; i<1000000; i++) MulMatrix4(&a, &b, &c);

  gettimeofday(&now, &tzone);
  elapsedTime = (double)(now.tv_sec - start.tv_sec) +
    (double)(now.tv_usec - start.tv_usec)/1000000.0;
  printf("%4.6f sec\n", elapsedTime);

  PrintMatrix4(&c);

  printf("start ASM version\n");
  gettimeofday(&start, &tzone);

  for(i=0; i<1000000; i++) MulMatrix4x4asm(&c, &a, &b);

  gettimeofday(&now, &tzone);
  elapsedTime = (double)(now.tv_sec - start.tv_sec) +
    (double)(now.tv_usec - start.tv_usec)/1000000.0;
  printf("%4.6f sec\n", elapsedTime);

  PrintMatrix4(&c);

  return 0;
}
