@----------------------------------------------------------
@ mat4x4.s   2012/08/19 Jun Mizutani 
@ as -mcpu=arm1176jzf-s -mfpu=vfp mat4x4.s -o mat4x4asm.o
@ gcc -O2 -Wall -o mat4 mat4.c mat4x4asm.o
@---------------------------------------------------------

.text
.global MulMatrix4x4asm

@ MulMatrix4x4asm(Matrix4 *result, Matrix4 *a, Matrix4 *b)
@ [result] = [a] x [b]
@ [  r0  ] = [r1] x [r2]
        .align  3
MulMatrix4x4asm:
        stmfd   sp!, {r0-r3, lr}
        vpush   {d0-d15}
        bl      vector4_mode            @ LEN=4: vector mode
        mov     r3, r0
        fldmiad r2!, {d0-d3}            @ vec_e
        mov     r0, r1                  @ vec_a
        fldmiad r0!, {d8-d11}           @ load a[0_3]
        fmuld   d4, d8, d0              @ w[i] =        a[i] * e0
        fldmiad r0!, {d12-d15}          @ load b[0-3]
        fmacd   d4, d12, d1             @ w[i] = w[i] + b[i] * e1
        fldmiad r0!, {d8-d11}           @ load c[0-3]
        fmacd   d4, d8, d2              @ w[i] = w[i] + c[i] * e2
        fldmiad r0, {d12-d15}           @ load d[0-3]
        fmacd   d4, d12, d3             @ w[i] = w[i] + d[i] * e3
        fstmiad r3!, {d4-d7}            @ store w[i]
        fldmiad r2!, {d0-d3}            @ vec_f
        mov     r0, r1                  @ vec_a
        fldmiad r0!, {d8-d11}
        fmuld   d4, d8, d0              @ x[i] =        a[i] * f0
        fldmiad r0!, {d12-d15}
        fmacd   d4, d12, d1             @ x[i] = x[i] + b[i] * f1
        fldmiad r0!, {d8-d11}
        fmacd   d4, d8, d2              @ x[i] = x[i] + c[i] * f2
        fldmiad r0, {d12-d15}
        fmacd   d4, d12, d3             @ x[i] = x[i] + d[i] * f3
        fstmiad r3!, {d4-d7}            @ store x[i]
        fldmiad r2!, {d0-d3}            @ vec_g
        mov     r0, r1                  @ vec_a
        fldmiad r0!, {d8-d11}
        fmuld   d4, d8, d0              @ y[i] =        a[i] * g0
        fldmiad r0!, {d12-d15}
        fmacd   d4, d12, d1             @ y[i] = y[i] + b[i] * g1
        fldmiad r0!, {d8-d11}
        fmacd   d4, d8, d2              @ y[i] = y[i] + c[i] * g2
        fldmiad r0, {d12-d15}
        fmacd   d4, d12, d3             @ y[i] = y[i] + d[i] * g3
        fstmiad r3!, {d4-d7}            @ store y[i]
        fldmiad r2, {d0-d3}             @ vec_h
        mov     r0, r1                  @ vec_a
        fldmiad r0!, {d8-d11}
        fmuld   d4, d8, d0              @ z[i] =        a[i] * h0
        fldmiad r0!, {d12-d15}
        fmacd   d4, d12, d1             @ z[i] = z[i] + b[i] * h1
        fldmiad r0!, {d8-d11}
        fmacd   d4, d8, d2              @ z[i] = z[i] + c[i] * h2
        fldmiad r0, {d12-d15}
        fmacd   d4, d12, d3             @ z[i] = z[i] + d[i] * h3
        fstmiad r3, {d4-d7}

        bl      scalar_mode             @ LEN=1: scalar mode
        vpop    {d0-d15}
        ldmfd   sp!, {r0-r3, pc}

@-------------------------------------------------------------------------
@ set FPSTR STRIDE[21:20] and LEN[18:16]
@-------------------------------------------------------------------------
scalar_mode:
        stmfd   sp!, {r2, lr}
        fmrx    r2, FPSCR              @ FPSCR --> r10
        bic     r2, r2, #0x00370000    @ clear (STRIDE=1 and LEN=1)
        fmxr    FPSCR, r2              @ r10 --> FPSCR
        ldmfd   sp!, {r2, pc}

vector4_mode:
        stmfd   sp!, {r2, lr}
        fmrx    r2, FPSCR              @ FPSCR --> r10
        bic     r2, r2, #0x00370000    @ clear (STRIDE=1 and LEN=1)
        orr     r2, r2, #0x00030000    @ set bit[17:16] (STRIDE=1 and LEN=4)
        fmxr    FPSCR, r2              @ r10 --> FPSCR
        ldmfd   sp!, {r2, pc}

