#!/usr/bin/luajit

-- -----------------------------
-- mulmatrix2.lua
--   2012/02/02 - Jun Mizutani
-- -----------------------------
ffi = require "ffi"

ffi.cdef[[
    typedef struct timeval {
        long tv_sec;
        long tv_usec;
    } timeval;
    int gettimeofday(struct timeval *tv, void *tz);
]]

function gettimeofday()
  local t = ffi.new("timeval")
  ffi.C.gettimeofday(t, nil)
  return t.tv_sec, t.tv_usec
end

function mulMatrix(mr, ma, mb)
  local a0 =ma[ 0]; local a1 =ma[ 1]; local a2 =ma[ 2]; local a3 =ma[ 3]
  local a4 =ma[ 4]; local a5 =ma[ 5]; local a6 =ma[ 6]; local a7 =ma[ 7]
  local a8 =ma[ 8]; local a9 =ma[ 9]; local a10=ma[10]; local a11=ma[11]
  local a12=ma[12]; local a13=ma[13]; local a14=ma[14]; local a15=ma[15]
  local b0 =mb[ 0]; local b1 =mb[ 1]; local b2 =mb[ 2]; local b3 =mb[ 3]
  local b4 =mb[ 4]; local b5 =mb[ 5]; local b6 =mb[ 6]; local b7 =mb[ 7]
  local b8 =mb[ 8]; local b9 =mb[ 9]; local b10=mb[10]; local b11=mb[11]
  local b12=mb[12]; local b13=mb[13]; local b14=mb[14]; local b15=mb[15]

  mr[ 0] = a0 * b0 + a4 * b1 +  a8 * b2 + a12 * b3
  mr[ 1] = a1 * b0 + a5 * b1 +  a9 * b2 + a13 * b3
  mr[ 2] = a2 * b0 + a6 * b1 + a10 * b2 + a14 * b3
  mr[ 3] = a3 * b0 + a7 * b1 + a11 * b2 + a15 * b3
  mr[ 4] = a0 * b4 + a4 * b5 +  a8 * b6 + a12 * b7
  mr[ 5] = a1 * b4 + a5 * b5 +  a9 * b6 + a13 * b7
  mr[ 6] = a2 * b4 + a6 * b5 + a10 * b6 + a14 * b7
  mr[ 7] = a3 * b4 + a7 * b5 + a11 * b6 + a15 * b7
  mr[ 8] = a0 * b8 + a4 * b9 +  a8 * b10+ a12 * b11
  mr[ 9] = a1 * b8 + a5 * b9 +  a9 * b10+ a13 * b11
  mr[10] = a2 * b8 + a6 * b9 + a10 * b10+ a14 * b11
  mr[11] = a3 * b8 + a7 * b9 + a11 * b10+ a15 * b11
  mr[12] = a0 * b12+ a4 * b13+  a8 * b14+ a12 * b15
  mr[13] = a1 * b12+ a5 * b13+  a9 * b14+ a13 * b15
  mr[14] = a2 * b12+ a6 * b13+ a10 * b14+ a14 * b15
  mr[15] = a3 * b12+ a7 * b13+ a11 * b14+ a15 * b15
end

function printMatrix4(ma)
  print( string.format(" %10.5e  %10.5e  %10.5e  %10.5e",
		       ma[0], ma[4], ma[8],  ma[12]))
  print( string.format(" %10.5e  %10.5e  %10.5e  %10.5e",
		       ma[1], ma[5], ma[9],  ma[13]))
  print( string.format(" %10.5e  %10.5e  %10.5e  %10.5e",
		       ma[2], ma[6], ma[10], ma[14]))
  print( string.format(" %10.5e  %10.5e  %10.5e  %10.5e",
		       ma[3], ma[7], ma[11], ma[15]))
end

local a = ffi.new("double[?]", 16, 1, 2, 3, 4,  1, 2, 3, 4,  
			       1, 2, 3, 4,  1, 2, 3, 4)
local b = ffi.new("double[?]", 16, 0, 1, 2, 3,  4, 5, 6, 7,
                             8, 9, 10, 11,  12, 13, 14, 15)
local c = ffi.new("double[?]", 16, 0, 0, 0, 0,  0, 0, 0, 0,  
                             0, 0, 0, 0,  0, 0, 0, 0)

local start_sec, start_usec = gettimeofday()

for i=1,1000000 do
  mulMatrix(c, a, b)
end

local end_sec, end_usec = gettimeofday()
local sec = end_sec - start_sec
local usec = end_usec - start_usec
if usec < 0 then 
    usec = usec + 1000000
    sec = sec - 1
end
print(string.format("luajit + ffi : %f sec", sec+usec/1000000))

printMatrix4(c)
